/*
 * File srvip.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL),
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file srvip.h
 *
 * server-side (and client-side) structs, macros, prototypes, etc.
 *
 */
#ifndef SRVIP_H_
#define SRVIP_H_

#ifndef NO_DOXYGEN
#include "project.def"
#include "srvcore.h"
#include "ttycmd.h"
#endif /* NO_DOXYGEN */

#if defined(UNIX) /* UNIX */
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <stddef.h>
#include <poll.h>
#endif

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/* some definitions for those stacks which don't support ipv6 */
#ifndef AF_IPX
# define AF_IPX 6
#endif
#ifndef IPV6_ADD_MEMBERSHIP
# define IPV6_ADD_MEMBERSHIP 12
# define IPV6_DROP_MEMBERSHIP 13
#endif
#ifndef USE_IPV6
# define AF_UNSPEC 0
# define AF_INET6 23
# define NI_NAMEREQD 0x04
# define IPPROTO_IPV4 4
# define IPPROTO_IPV6 41
# define IPV6_MULTICAST_HOPS 18
# define AI_PASSIVE 0x00000001
# define AI_CANONNAME 0x00000002
# define ss_family sin_family
struct in6_addr
{
  union
  {
    BYTE u6_addr8[16];
    UINT16 u6_addr16[8];
    UINT32 u6_addr32[4];
  } in6_u;
# define s6_addr  in6_u.u6_addr8
# define s6_addr16  in6_u.u6_addr16
# define s6_addr32  in6_u.u6_addr32
};
typedef struct in6_addr IN6_ADDR;
extern IN6_ADDR in6addr_any;
extern IN6_ADDR in6addr_loopback;
struct sockaddr_in6
{
  SINT16 sin6_family;
  SINT16 sin6_port;
  UINT32 sin6_flowinfo;
  struct in6_addr sin6_addr;
  UINT32 sin6_scope_id;
};
struct addrinfo
{
  int ai_flags;            /* AI_PASSIVE, AI_CANONNAME */
  int ai_family;           /* AF_UNSPEC, AF_INET, AF_INET6 */
  int ai_socktype;         /* SOCK_STREAM, SOCK_DGRAM */
  int ai_protocol;         /* IPPROTO_IP, IPPROTO_IPV6 */
  size_t ai_addrlen;       /* length of ai_addr */
  struct sockaddr *ai_addr;/* socket address struct */
  char *ai_canonname;     /* canonnical name */
  struct addrinfo *ai_next;/* next pointer */
};

typedef struct ipv6_mreq
{
  struct in6_addr ipv6mr_multiaddr;  /* IPv6 multicast address */
  unsigned int    ipv6mr_interface;  /* Interface index */
} IPV6_MREQ;

int getaddrinfo(const char *hostname,
                const char *service,
                const struct addrinfo *hints,
                struct addrinfo **res);
void freeaddrinfo(struct addrinfo *ai);
int getnameinfo(const struct sockaddr *sa, _socklen_t salen,
                       char *host, size_t hostlen,
                       char *serv, size_t servlen, int flags);
int inet_pton(int family,const char *src,void *dst);
char *inet_ntop(int family,const void *src,char *dst,size_t cnt);
# define SCKADR_LEN 4
#else
# define SCKADR_LEN 16
# if defined(WIN32)
# include <iphlpapi.h>
# elif defined(UNIX)
# include <net/if.h>
# ifndef SOLARIS
# include <ifaddrs.h>
# endif
# ifdef FREEBSD
#  include <netdb.h>
#  include <net/if_dl.h>
# endif
# endif
#endif /* !defined(USE_IPV6) */
extern int gAfInet;
extern int gInAdrLen;
extern int gUseLoopback;
extern int gNoLoopback;
extern INADR inadr_zero;
#define adrcmp(a,b) memcmp(a,b,gInAdrLen)
#define adrcpy(a,b) memset(a,0,sizeof(INADR)), memcpy(a,b,gInAdrLen)
#define sa4_eq(a,b) (((a)->sin_port == (b)->sin_port) && !memcmp(&(a)->sin_addr,&(b)->sin_addr,4))
#define sa6_eq(a,b) (((a)->sin6_port == (b)->sin6_port) && !memcmp(&(a)->sin6_addr,&(b)->sin6_addr,16))
#define sckadr_eq(a,b) (\
 ((a)->ss_family == (b)->ss_family) &&\
 ((a)->ss_family == AF_INET6 ?\
 sa6_eq((struct sockaddr_in6 *)a,(struct sockaddr_in6 *)b) :\
 sa4_eq((struct sockaddr_in *)a,(struct sockaddr_in *)b))\
)
# define setmcopts(l,o) {\
if (gAfInet == AF_INET6) {\
  l = IPPROTO_IPV6; o = IPV6_MULTICAST_HOPS;\
} else {\
  l = IPPROTO_IP; o = IP_MULTICAST_TTL;\
} }

TINE_EXPORT int SetAfInet(int family);
TINE_EXPORT int GetAfInet(void);
/**
 * \brief Gets the address (ip address and port) of the given socket address
 * as a string.
 *
 * \param sckadr is a reference to a socket address whose string representation is desired.
 * \param addr is a string buffer with enough space to hold the string representation
 * of the given socket address (e.g. at least 64 characters)
 *
 * \return 0 or a TINE error code if the call fails.
 *
 */
TINE_EXPORT int GetInetAddress(SCKADR *sckadr, char *addr);
/**
 * \brief Gets the host name from the address given.
 *
 * \param addr is ip address whose host name is desired.
 * \param host is a string buffer which will contain the host name when the call completes.
 * \param length is the length of the string buffer passed as the 'host' parameter.
 *
 * \return 0 or a TINE error code if the call fails.
 */
TINE_EXPORT int GetHostNameFromAddress(char *addr, char *host, int length);
int getInetFamily(int inetProtocol);
int inetStrAdr(SCKADR *sckadr,char *dst,int addport);
UINT16 inetPort(SCKADR *sckadr);
void setInAddrAny(SCKADR *sckadr);
int isInAddrAny(SCKADR *sckadr);
void setInAddrLoopback(SCKADR *sckadr);
int isInAddrLoopback(SCKADR *sckadr);
BYTE *getInAddr(SCKADR *sckadr);
int getInAddrLen(int family);
int sckadrcmp(SCKADR *adr1,SCKADR *adr2);
void rprtSckErr(const char *src,const char *txt,SOCKET sck,int err, int debug,char *dbgbuf);
char *getSckStr(SOCKET sck);
#define IPV6_IPV4PREFIX "::ffff:"
char *toIpv4Addr(char *ipv6addr,char *ipv4addr);
char *toIpv6Addr(char *ipv4addr,char *ipv6addr);
int isIpv4on6(char *addr);
int fixInetProtocol(FecAddrStruct *fec,int protocol);
extern int gSufferInSilence;
TINE_EXPORT void SetSufferInSilence(int value);

#ifndef NumTCPChan
# define NumTCPChan	5
#endif
#ifndef UDP_PORT
# define UDP_PORT 9003
#endif
extern int gtUDPPort;
#ifndef TCP_PORT
# define TCP_PORT 8503
#endif
extern int gtTCPPort;
#ifndef GCAST_PORT
# define GCAST_PORT 8004
#endif
extern int gtGCastPort;
#ifndef NETSRV_PORT
# define NETSRV_PORT 9002
#endif
extern int gtNetSrvPort;
#ifndef MCAST_PORT
# define MCAST_PORT 9503
#endif
extern int gtMCastPort;
#ifndef STREAM_PORT
# define STREAM_PORT 10003
#endif
extern int gtStrmPort;
#ifndef PIPE_PORT
# define PIPE_PORT 10503
#endif
extern int gtPipePort;
#ifndef UDP_CLN_PORT
# define UDP_CLN_PORT 8050
#endif
extern int gtUDPClnPort;
#ifndef UDP_CLN_RANGE
# define UDP_CLN_RANGE (UDP_CLN_PORT+1000)
#endif
#ifndef NETBCST_ADDR
# define NETBCST_ADDR "0.0.0.0"
#endif
extern int gMultiCastUseLegacyRules;
extern int gDcsMultiCastUseLegacyRules;
extern char *gtNetLclBcast;
#ifndef NETSRV_ADDR
# define NETSRV_ADDR "239.1.1.2"
#endif
#ifndef NETSRV_ADDR6
# define NETSRV_ADDR6 "ff05::0:0:0:0:1:1:2"
#endif
#define NETSRV_ADDR_LEGACY "238.1.1.2"
extern char *gtNetSrvAddr;
#ifndef MCAST_ADDR
# define MCAST_ADDR "239.1.1.1"
#endif
#ifndef MCAST_ADDR6
# define MCAST_ADDR6 "ff05::"
#endif
#define MCAST_ADDR_LEGACY "238.1.1.1"
extern char *gtMCastAddr;
#ifndef GCAST_ADDR6
# define GCAST_ADDR6 "ff05::"
#endif
#ifndef GCAST_ADDR
# define GCAST_ADDR "239.1.1.0"
#endif
#define GCAST_ADDR_LEGACY "238.1.1.0"
extern char *gtMCastAddrLegacy;
#ifndef MCAST_TTL
# define MCAST_TTL 16
#endif
extern int gtMCastTTL;
#ifndef IPC_PATH
# define IPC_PATH "/tmp/"
#endif
#define IPC_PERM (S_IRWXU|S_IRWXG|S_IRWXO)

#ifndef MAX_TCPCONNECTIONS
# define MAX_TCPCONNECTIONS 32
#endif

#ifndef MAX_IPCCONNECTIONS
# define MAX_IPCCONNECTIONS 64
#endif

#ifndef MAX_DBGCONNECTIONS
# define MAX_DBGCONNECTIONS 5
#endif

#ifndef DEFAULT_MAX_TCPMSG_SIZE
# define DEFAULT_MAX_TCPMSG_SIZE 100000000
#endif

#define MIN_FDSETS 16

#define MTU 1460

#ifndef MIN_RCVBUF
/**
 * \internal
 * \brief Default minimal number of receive buffers for socket options.
 */
# define MIN_RCVBUF 0x1000
#endif

#ifndef MAX_RCVBUF
/**
 * \internal
 * \brief Default maximal number of receive buffers for socket options.
 */
# define MAX_RCVBUF 0x10000
#endif

#ifndef MIN_SNDBUF
/**
 * \internal
 * \brief Default minimal number of send buffers for socket options.
 *
 */
# define MIN_SNDBUF MIN_RCVBUF
#endif

#ifndef MAX_SNDBUF
/**
 * \internal
 * \brief Default maximal number of send buffers for socket options.
 *
 */
# define MAX_SNDBUF (MAX_RCVBUF / 2)
#endif

#ifndef DEFAULT_TCP_CONNECTION_TIMEOUT
# define DEFAULT_TCP_CONNECTION_TIMEOUT 200
#endif

#ifndef TIME_SERVER_HOST
# define TIME_SERVER_HOST NULL
#endif

#define LOOPBACKADDR_IPV4 "127.0.0.1"
#define LOOPBACKADDR_IPV6 "::1"
extern char *gLoopbackAddr;

#ifndef IPPORT_ECHO
# define IPPORT_ECHO 7
#endif

#if defined(NEED_POLL)
# ifndef ECONNABORTED
# define ECONNABORTED    106
# endif
# define POLLIN          0x0001
# define POLLPRI         0x0002
# define POLLOUT         0x0004
# define POLLERR         0x0008
# define POLLHUP         0x0010
# define POLLNVAL        0x0020
struct pollfd
{
  int   fd;         /* file descriptor */
  short events;     /* requested events */
  short revents;    /* returned events */
};
int poll(struct pollfd *fds, int nfds, int timeout);
#elif defined(WIN32)
# define poll WSAPoll
#endif /* NEED_POLL */
int getFdsIdx(struct pollfd *fds,int nfds,SOCKET s);
struct pollfd *getFds(struct pollfd *fds,int nfds,SOCKET s);
int fdsContains(struct pollfd *fds, int nfds, SOCKET s);
#define ADDSCKTOSET(s) if ((s) > 0 && n < maxlen) {fds[n].fd = (s); fds[n].events = POLLIN; n++;}
#define fds_isset(s) ((idx=getFdsIdx(fds,nfds,(s))) >= 0 && (fds[idx].revents & POLLIN) != 0)
/* n.b. the following macro can only be used directly after fds_isset*/
#define fds_iserr(s) (idx > 0 && (fds[idx].revents & (POLLERR|POLLHUP|POLLNVAL)))
#define pollerr(i) (fds[i].revents & (POLLERR|POLLHUP|POLLNVAL))
#define pollhup(i) (fds[i].revents & (POLLHUP))
#define EXT_POLLFDS_CAPACITY 100
#define pollerr2errno(e) ((e) & POLLHUP ? ECONNABORTED : (e) & POLLERR ? EIO : EBADF)
extern struct pollfd *ext_pollfds;
extern int next_pollfds;

/* PollExternalDescriptor() */
/**
 * \brief registers an external socket to the server's poll set
 *
 * TINE attempts will call a blocking 'poll' on the known socket descriptors.
 * If the application needs to react to an event on an external socket on the
 * same thread as the tine cycler, such sockets can be registered with the
 * cycler via this call.
 *
 * \param s is the external socket descriptor to be added to the poll set.
 * \param events is the requested events for which the 'poll' should wake up.
 *
 * \return 0 on success or a TINE error code
 */
TINE_EXPORT int PollExternalDescriptor(SOCKET s,int events);

#define NETSRV_SOCKET 6565

#define MY_ADDRESS_FILE "myaddr.csv"

#define MCAST_ADDRESS_FILE "mcastaddr.csv"
#define MCAST_MASK_FILE "mcastmask.csv"
#define GCAST_ADDRESS_FILE "gcastaddr.csv"
#define GCAST_MASK_FILE "gcastmask.csv"

#define ADDRFLAG_MCAST_LEGACY 0x00
#define ADDRFLAG_MCAST_MODERN 0x01
#define ADDRFLAG_LOCAL_HOST   0x10
#define ADDRFLAG_LOCAL_PROCESS 0x20

#define ADDRFLAG_BIGENDIAN   0x100

typedef struct
{
  INADR mcInAdr; /**< assigned multicast INADR */
  INADR patInAdr;/**< address pattern as INADR */
  char mcStrAdr[ADDR_SIZE_IPV6]; /**< assigned multicast address */
  char patStrAdr[ADDR_SIZE_IPV6]; /**< address pattern */
} MCastAddrMap;
extern MCastAddrMap *gtMCastAddrLst;
extern int nMCastAddrLstEntries;
INADR fixMcastHaddr(BYTE *bAdr,int legacyRules);
int isInvalidIpV4Addr(char *addr);
int isInvalidIpV6Addr(char *addr);
int dumpMCastAddrList(void);

#define NETS_ACCESS  0
#define NETS_NETWORK 1
#define NETS_DENIED 2
extern char *netsListFile[3];

typedef struct tcpIdleConnectionStruct
{
  int idx; /**< FEC table index */
  SOCKET sck; /**< TCP socket */
  int trp; /**< transport (TCP or STREAM) */
  int ttl; /**< time to live */
  struct tcpIdleConnectionStruct *prv;
  struct tcpIdleConnectionStruct *nxt;
} IDLE_CONNECTION;
extern IDLE_CONNECTION *tcpIdleConnectionList;
IDLE_CONNECTION *GetIdleConnection(int idx,int inetProtocol);
IDLE_CONNECTION *AddIdleConnection(SOCKET sck,int idx,int inetProtocol);
IDLE_CONNECTION *RemoveIdleConnection(IDLE_CONNECTION *ic);
IDLE_CONNECTION *GetIdleConnectionBySocket(SOCKET sck,int inetProtocol);
void rmvIdleConnections(void);

/**
 * \internal
 * \brief stream parameters used for TCP Stream transport
 *
 */
typedef struct
{
  ClnHdr dst;
  BYTE *payload;
  int length;
  int capacity;
} StreamParams;
int putStreamParams(ClnHdr *c,BYTE *payload,UINT32 msgsize,int mark);

#define SD_THREAD_NONE 0
#define SD_THREAD_CREATED 1
#define SD_THREAD_RUNNING 2
#define SD_THREAD_STOPPING 3
#define SD_THREAD_STOPPED 4
typedef struct
{
  ClnHdr dst;
  BYTE payload[TCP_BUFFER_SIZE];
  int msgsize;
  int isready;
  int isbusy;
} StreamDataTable;
typedef struct tcpbckt
{
  int idx;                  /**< FEC index             */
  SOCKET sck;               /**< TCP connection socket */
  int inetprot;             /**< inet protocol (TCP, TCP6, ...) */
  int tineprot;             /**< tine protocol level */
  int ptr;                  /**< bucket pointer        */
  SINT32 siz;               /**< bucket target size    */
  int len;                  /**< bucket buffer length  */
  int clrlen;               /**< collector buffer length */
  int errcnt;               /**< error count on this socket */
  int rmvNeeded;            /**< flagged removal */
  BYTE *buf;                /**< bucket buffer         */
  BYTE *clrbuf;             /**< collector buffer      */
# ifdef MULTITHREADED
  THRHANDLE rcvThr;         /**< recvThread handle     */
  THRHANDLE sndThr;         /**< sendThread handle     */
  SEMHANDLE sndSem;         /**< sendThread semaphore  */
# ifdef UNIX
  pthread_mutex_t psndMtx;  /**< primary sendThread mutex (posix builds) */
# endif
  MXHANDLE sndMtx;          /**< sendThread mutex handle */
  StreamParams sndPrm;      /**< stream parameters for STREAM transfers */
  int sdTblHead;            /**< stream data table head */
  int sdTblTail;            /**< stream data table tail */
  int sdTblMark;            /**< stream data mark start of next payload */
  int sdTblSize;            /**< stream data table size */
  int sdTblWaitMark;        /**< waiting for next mark */
  int sdThrState;           /**< steam data thread state */
  StreamDataTable *sdTbl;   /**< stream data table (queue) */
# endif
  struct tcpbckt *nxt;      /**< this is a linked list */
} TCPBCKT;
extern TCPBCKT *bcktList;
TCPBCKT *findBucket(SOCKET sck,int lock);
TCPBCKT *getBucket(SOCKET sck,int buffer_length,int collector_length,int inetprot,int tineprot, int source);
void freeBucket(SOCKET sck,int yield);
void freeBucketThreadMemory(TCPBCKT *bckt);
void removeBucketFromList(TCPBCKT *bckt);

/**
 * \internal
 * \brief used as a table and/or linked list
 */
typedef struct TcpSckT
{
  SOCKET sck;
  int type;
  int idx;
  struct TcpSckT *nxt;
} TCPSCKT;
extern TCPSCKT *tcpSckTbl;
extern int nTCPsck;
extern int tcpConTblSize;
extern int gMaxTcpMsgSize;

extern TCPSCKT *dbgSckTbl;
extern int nDBGsck;
extern int dbgConTblSize;
extern int gAllowRemoteDebugging;

extern TCPSCKT *tcpClnSckTbl;
TCPSCKT *getTcpClnSck(SOCKET sck, int index, int type);
void rmvTcpClnSck(SOCKET sck);
void rmvTcpClnSckList(void);

extern UINT32 sendtoErrcnt;
extern UINT32 selectErrcnt;
extern UINT32 selectFailcnt;

struct dcsEvnPktHdr
{
  UINT32 tag;    /**< magic number */
  UINT32 grp;    /**< source multicast IP address */
  UINT32 sec;    /**< acquisition time (sec) */
  UINT32 usec;   /**< acquisition time (usec) */
  UINT32 evnId;  /**< event */
  UINT32 blkId;  /**< server block ID */
  UINT16 pktNum; /**< packet number */
  UINT16 numPkt; /**< number of packets in server block */
  UINT16 datLen; /**< length of data in the packet */
  UINT16 sts;    /**< status */
  UINT32 totLen; /**< total length of data block */
};

extern int ipxSrvSck;
extern int ipxClnSck;
extern int ipxMcastClnSck;
extern int ipcSrvSck;
extern int gBadPacketCnts;
#define soerr(s) {cc = udp_socket_error; strcat(txt,s); goto err;}
#define addIpSubStr(b,s) if ((s).mask != 32) sprintf(&(b)[strlen(b)],"/%d",(s).mask)
#define IsConnProt(p) ((p) == TCP || (p) == STREAM)

extern EpAdr gInAdr;

extern int gSndIpBroadcastsOnly;
extern INADR tMCastAddr;
extern INADR tMCastAddrLegacy;
extern INADR tGCastAddrLegacy;
extern INADR tNetSrvAddr;
extern SCKADR tGCastSckAddr;
extern fd_set ext_set;
extern int externalFdSets;
extern SCKACL *IpBcstLst;
extern int sizIpBcstLst;
void FD_MERGE(fd_set fdsrc,fd_set *fddst);
int addServerSocketsToSet(struct pollfd *fds,int offset,int maxlen);
int addClientSocketsToSet(struct pollfd *fds,int offset,int maxlen);
extern void (*socketAcceptEntryFcn)(void);
int setupUdpSocket(SOCKET *s,int family,unsigned short port,int initMultiCast,int rcvbuf,int sndbuf);
void CloseConnectionSocket(SOCKET sck,int inetProtocol,int yield);
void closeIPConnectionToFec(int index,SOCKET sck);
int ListenForIPMulticasts(void);
int ListenForIPNetGlobals(void);
int ListenForDcsEvns(BYTE *haddr);
BYTE *getmyipaddr(void);
char *getmyipaddrstr(void);
int GetThisFecAddresses(FecAddrStruct *fec);
int applyBcstAddr(int family,INADR *src,INADR *dst);
int getSrvIPAddr(void);
int getRegNetsFromFile(const char *eqm,const char *dev,const char *prp,const char *filename,int type,int *nr,SCKACL **list);
int addRegNetsLst(const char *filename,int type,int *nr,SCKACL **list,NAME64 *ipaddr,int addsiz);
int rmvRegNetsLst(const char *filename,int type,int *nr,SCKACL **list,NAME64 *ipaddr,int addsiz);
int attachPortToMulticastGroup(SOCKET sck,UINT16 port,INADR grpAddr,char *grpType);
int detachPortFromMulticastGroup(SOCKET sck,UINT16 port,INADR grpAddr,char *grpType);
void freeBucketList(void);
void checkSckBufferSpace(SOCKET sck,int size,int opt,char *tag);
int ip4adr(SCKADR *sckadr,BYTE *adr,UINT16 port);
int ip6adr(SCKADR *sckadr,BYTE *adr,UINT16 port);
int sckadr_in(SCKADR *sckadr,BYTE *adr,UINT16 port,int family);
void removeFlaggedBuckets(void);

extern SCKACL *denynets;
extern int ndenynets;

TINE_EXPORT char *GetMyIpAddress(void);
TINE_EXPORT int SetSrvUDPPort(int port);
TINE_EXPORT int GetSrvUDPPort(void);
TINE_EXPORT int SetSrvTCPPort(int port);
TINE_EXPORT int GetSrvTCPPort(void);
TINE_EXPORT int SetGCastPort(int port);
TINE_EXPORT int GetGCastPort(void);
TINE_EXPORT int SetMCastPort(int port);
TINE_EXPORT int GetMCastPort(void);
TINE_EXPORT int OpenIpcPipe(char *pipeName);
TINE_EXPORT int openIPCPipe(char *pipeName);
TINE_EXPORT int CloseIpcPipe(char *pipeName);
TINE_EXPORT int GetIpcPipeName(char *pipeName);

/* SetClnUDPPort() */
/**
 * \brief Sets the TINE client UDP base port
 *
 * TINE attempts to obtain the first free port beginning UDP client base port.
 * If for some reason the client base port needs to be adjusted away from its default
 * value of 8050, then this routine can be used.
 *
 * \param port is the base UDP port to use for the TINE clients
 *
 * \return 0 on success or a TINE error code
 */
TINE_EXPORT int SetClnUDPPort(int port);

/* GetClnUDPPort() */
/**
 * \brief Gets the TINE client UDP base port
 *
 * \return The client base port
 */
TINE_EXPORT int GetClnUDPPort(void);

/* SetMCastTTL() */
/**
 * \brief Sets the TINE multicast Time-To-Live
 *
 * If the multicast Time-To-Live parameter (number of 'hops') needs to be
 * adjusted away from its default value of 16 then this routine can be used.
 *
 * \param ttl is the time-to-live value to use for multicasts
 *
 * \return 0 on success or a TINE error code
 */
TINE_EXPORT int SetMCastTTL(int ttl);

/* GetMCastTTL() */
/**
 * \brief Gets the TINE multicast Time-To-Live
 *
 * \return The multicast Time-To-Live
 */
TINE_EXPORT int GetMCastTTL(void);

/* SetPacketMTU() */
/**
 * \brief Sets the TINE Packet MTU for the UDP server socket
 *
 * The TINE default transport protocol is UDP and the default UDP datagram
 * MTU (maximum transport unit) is 1472 bytes (i.e. the ethernet packet MTU).
 * TINE will supply its own packet reassembly for data transfers larger than this.
 * Many modern operating systems have network stacks that will provide
 * UDP packet reassembly up to 64 Kbytes. This generally a much more efficent
 * packet reassembly as it happens at the driver level.  If a higher performance
 * for very large data transfer is required, this routine can be called at the
 * server prior to any call to SystemInit().
 *
 * \param mtu is the desired maximum transport unit for the server socket.
 * The default is 1472 bytes.  The minimum accepted value is 512 and the
 * maximum accepted value is 64000 bytes.
 *
 * \return 0 on success or a TINE error code
 *
 * \note Not all operating systems are able to handle MTUs greater than the standard
 * value of 1472 (notably the standard VxWorks build, and some legacy operating systems).
 * Clients on these operating systems will not be able to receive large data transfers
 * from servers sending at greater MTUs!
 */
TINE_EXPORT int SetPacketMTU(int mtu);

/* GetPacketMTU() */
/**
 * \brief Gets the TINE UDP server packet MTU
 *
 * \return The UDP server packet MTU
 */
TINE_EXPORT int GetPacketMTU(void);
TINE_EXPORT int SetNetSrvAddr(char *addr);
TINE_EXPORT char *GetNetSrvAddr(void);
/* SetMCastAddr() */
/**
 * \brief Sets the cannonical multicast address the server will use when
 * requested to send payloads via multicast.
 *
 * A TINE server will by default use the multicast address '239.b.c.d' where
 * bytes 'b', 'c', and 'd' are obtained from the server's own IP address

 * '238.1.c.d.' where
 * the last 2 bytes 'c' and 'd' are obtained from the server's own IP address
 * (in IPv4 notation).  Under some circumstances it might be desireable to make
 * use of a specific multicast address (e.g. for failover of a globals variable).
 * The call can be used for this purpose.  An alternative is to utilize the
 * environment variable 'TINE_MCAST_ADDR' or to use the configuration file:
 * 'mcastaddr.csv'.  However in such cases, care should be
 * taken to ensure that only the server application sees such an environment variable
 * or such a configuration file.
 *
 * \param addr is the desired multicast address for the server.  In IPv4 this
 * must be in the form of 'a.b.c.d'.
 *
 * \return 0 on success or a TINE error code
 *
 */
TINE_EXPORT int SetMCastAddr(char *addr);
/**
 * \brief Sets the cannonical legacy multicast address the server will use when
 * where necessary.
 *
 * A TINE server prior to release 4.5.2 will by default use the multicast
 * address '238.1.c.d.' where
 * the last 2 bytes 'c' and 'd' are obtained from the server's own IP address
 * (in IPv4 notation).  Under some circumstances it might be desireable to make
 * use of a specific multicast address (e.g. for failover of a globals variable).
 * The call can be used for this purpose.  An alternative is to utilize the
 * environment variable 'TINE_MCAST_ADDR' or to use the configuration file:
 * 'mcastaddr.csv'.  However in such cases, care should be
 * taken to ensure that only the server application sees such an environment variable
 * or such a configuration file.
 *
 * \param addr is the desired multicast address for the server.  In IPv4 this
 * must be in the form of 'a.b.c.d'.
 *
 * \return 0 on success or a TINE error code
 *
 */
TINE_EXPORT int SetMCastAddrLegacy(char *addr);
/* GetMCastAddr() */
/**
 * \brief Gets the cannonical multicast address a server will use when
 * requested to send payloads via multicast.
 *
 * \return the server's canonical multicast address in string representation.
 *
 */
TINE_EXPORT char *GetMCastAddr(void);
/**
 * \brief Gets the cannonical legacy multicast address a pre-release 4.5.2 server will use when
 * requested to send payloads via multicast.
 *
 * \return the server's canonical multicast address in string representation.
 *
 */
TINE_EXPORT char *GetMCastAddrLegacy(void);
/* SetGCastAddr() */
/**
 * \brief Sets the cannonical globals multicast address the server will use when
 * sending payloads via multicast as a producer.
 *
 * A TINE server will by default use the multicast address '238.1.c.d.' where
 * the last 2 bytes 'c' and 'd' are obtained from the server's own IP address
 * (in IPv4 notation).  Under some circumstances it might be desireable to make
 * use of a specific multicast address (e.g. for failover of a globals variable).
 * The call can be used for this purpose.  An alternative is to utilize the
 * environment variable 'TINE_GCAST_ADDR' or to use the configuration file:
 * 'gcastaddr.csv'.  However in such cases, care should be
 * taken to ensure that only the server application sees such an environment variable
 * or such a configuration file.
 *
 * \param addr is the desired multicast address for the server.  This
 * must be in the form of 'a.b.c.d' (IPv4) or 'a:b:c:d::e:f:g:h' (IPv6).
 *
 * \return 0 on success or a TINE error code
 *
 */
TINE_EXPORT int SetGCastAddr(char *addr);
/* GetGCastAddr() */
/**
 * \brief Gets the cannonical globals multicast address the server will use when
 * sending payloads via multicast as a producer.
 *
 * \return the server's canonical globals multicast address in string representation.
 *
 */
TINE_EXPORT char *GetGCastAddr(void);
/* SetMCastMask() */
/**
 * \brief Sets the cannonical multicast address mask the server will use when
 * requested to send payloads via multicast.
 *
 * Under legacy conditions (gMultiCastUseLegacyRules = TRUE)
 * a TINE server will by default use the multicast address '238.1.c.d.' where
 * the last 2 bytes 'c' and 'd' are obtained from the server's own IP address
 * (in IPv4 notation).  The initial two bytes consitute the 'mask'.  The first
 * byte signals a multicast if it falls between 224 and 239.  The second byte
 * can be used as a multicast 'metric' to offer further distinction amoung
 * multicast senders if required.  This call can be used to specify the
 * multicast mask.  An alternative is to utilize the
 * environment variable 'TINE_MCAST_MASK' or to use the configuration file:
 * 'mcastmask.csv'.
 *
 * Under modern conditions
 * a TINE server will by default use the multicast address '239.b.c.d.' where
 * the last 3 bytes 'b', 'c', and 'd' are obtained from the server's own IP address
 * (in IPv4 notation).  The initial byte consitutes the 'mask' in this case.
 *
 * \param addr is the desired multicast address for the server.  This
 * must be a string in the form of 'mask1|pattern1,mask2|pattern2,...' where
 * the 'mask' and 'pattern' substrings must be of the form 'a.b.c.d' (IPv4)
 * or 'a:b:c:d::e:f:g:h' (IPv6).
 * If 'pattern' is not given, then the ip address of the local host is used.
 *
 * \return 0 on success or a TINE error code
 *
 */
TINE_EXPORT int SetMCastMask(char *addr);
/* GetMCastMask() */
/**
 * \brief Gets the cannonical multicast address mask the server will use when
 * requested to send payloads via multicast.
 *
 * \param pattern is the target ip address pattern for which the assigned
 * multicast mask is desired.  If 'pattern' is NULL then the canonical multicast
 * mask is returned.
 *
 * \return the server's canonical multicast address mask in string representation.
 *
 */
TINE_EXPORT char *GetMCastMask(char *pattern);
/* SetGCastMask() */
/**
 * \brief Sets the cannonical multicast address mask the server will use when
 * sending payloads via multicast as a producer.
 *
 * Under legacy conditions (gMultiCastUseLegacyRules = TRUE)
 * a TINE server will by default use the multicast address '238.1.c.d.' where
 * the last 2 bytes 'c' and 'd' are obtained from the server's own IP address
 * (in IPv4 notation).  The initial two bytes consitute the 'mask'.  The first
 * byte signals a multicast if it falls between 224 and 238.  The second byte
 * can be used as a multicast 'metric' to offer further distinction amoung
 * multicast senders if required.  This call can be used to specify the
 * multicast mask.  An alternative is to utilize the
 * environment variable 'TINE_GCAST_MASK' or to use the configuration file:
 * 'gcastmask.csv'.
 *
 * Under modern conditions
 * a TINE server will by default use the multicast address '239.b.c.d.' where
 * the last 3 bytes 'b', 'c', and 'd' are obtained from the server's own IP address
 * (in IPv4 notation).  The initial byte consitutes the 'mask' in this case.
 *
 * \param addr is the desired multicast address for the server.  In IPv4 this
 * must be in the form of 'a.b.c.d'.
 *
 * \return 0 on success or a TINE error code
 *
 */
TINE_EXPORT int SetGCastMask(char *addr);
/* GetGCastMask() */
/**
 * \brief Gets the cannonical multicast address mask the server will use when
 * sending payloads via multicast as a producer.
 *
 * \param addr is the desired multicast address for the server.  In IPv4 this
 * must be in the form of 'a.b.c.d'.
 *
 * \return the server's canonical multicast address mask in string representation.
 *
 */
TINE_EXPORT char *GetGCastMask(char *addr);

/* SetUseMCastLegacyRules() */
/**
 * \brief Establishes legacy mutlicast systematics or not
 *
 * \param value determines whether legacy multicast rules are in effect.
 *  If TRUE then the multicast byte = 238 and the second byte is
 *  '1' by default and the last 2 bytes are those of the sender.
 *  If FALSE (default) the multicast byte = 239 and the last
 *  3 bytes are those of the sender.
 *
 */
TINE_EXPORT void SetUseMCastLegacyRules(int value);

/* GetUseMCastLegacyRules() */
/**
 * \brief Returns whether legacy mutlicast systematics are in effect or not
 *
 * \return Whether legacy mutlicast systematics are in effect (TRUE) or not (FALSE).
 *
 */
TINE_EXPORT int GetUseMCastLegacyRules(void);

/* SetMaxTcpMessageSize() */
/**
 * \brief Sets the TINE server maximum STREAM message segment size
 *
 * When using CM_STREAM to acquire data from a server, very large amounts of
 * data can be passed to the stream socket.  In general the network stack will not
 * be able to prcess an indeterminantly large amount and the TINE transport system
 * will need to parcel the transmission into segments.  The
 * maximum tcp message size determines the size of the underlying parcel.
 * By default, this is 100000000 bytes.
 * This value can be adjusted by calling this routine at the server prior to any
 * call to SystemInit().
 *
 * \param value is the desired maximum STREAM message segment size
 * The default is 100000000 bytes.  The minimum accepted value is 1472 byte, There
 * is no maximum, however the network stack may refuse to trasmit if the value is too
 * large.
 *
 * \return 0 on success or a TINE error code
 *
 * \note this applies only to trasmissions in CM_STREAM mode, which is in turn only
 * applicable to multi-threaded builds of the TINE libraries.
 */
TINE_EXPORT void SetMaxTcpMessageSize(int value);

/* GetMaxTcpMessageSize() */
/**
 * \brief Gets the TINE server maximum STREAM message segment size
 *
 * \return The the TINE server maximum STREAM message segment size
 */
TINE_EXPORT int GetMaxTcpMessageSize(void);

/* _SystemAssignBufferSpace() */
/**
 * \brief Assigns specific buffer capacity for certain configurable TINE sockets.
 *
 * The standard TINE client and server sockets (both UDP, TCP, and STREAM) are initiated with a
 * reasonable default amount of receive (64 Kbyte) and send (32 Kbyte) buffer capacity.  If for any
 * performance reason, more or less buffer capactiy is needed, this routine should be called prior to
 * any calls to SystemInit() or SystemCycle().  For example, VxWorks builds which do not have a large
 * amount of network resources might want to set these values smaller than the defaults.  Video system
 * clients expecting extreme amounts of incoming data might want to set these values to a much larger
 * amount.
 *
 * The principal benificiaries of the 'receive buffer space' are the client sockets (UDP,TCP and SREAM), whereas
 * the principal benificiaries of the 'send buffer space' are the server sockets (TCP and STREAM). The
 * remaining socket buffer space (e.g. client-side send buffer space or server side receive buffer space)
 * will be held at 'reasonable' values, the assumption being that the dominant data delivery will generally
 * server (sending) to client (receiving).
 *
 * If you need additional control over socket settings, consider using the extended routine
 * _SystemAssignBufferSpaceEx().
 *
 * \param rcvBufferSpace is the amount of receive buffer space (in bytes) to assign to the configurable TINE sockets.
 * A value of '0' will assign the default receive buffer capacity.
 * \param sndBufferSpace is the amount of send buffer space (in bytes) to assign to the configurable TINE TCP or
 *  STREAM sockets.  The UDP socket send buffer space is not influenced by this parameter and
 * will be at least as large as twice the assigned UDP packet MTU. See _SystemAssignUdpSendBufferSpace().
 * A value of 0' will assign the default send buffer capacity.
 *
 * \return 0 on success or a TINE error code
 *
 * \see SetPacketMTU(), _SystemAssignBufferSpaceEx()
 */
TINE_EXPORT int _SystemAssignBufferSpace(UINT32 rcvBufferSpace,UINT32 sndBufferSpace);
#define SystemAssignBufferSpace(a,b) _SystemAssignBufferSpace(a,b)

/* _SystemAssignBufferSpaceEx() */
/**
 * \brief Assigns specific buffer capacity for certain configurable TINE sockets (extended version).
 *
 * The standard TINE client and server sockets (both UDP, TCP, and STREAM) are initiated with a
 * reasonable default amount of receive (64 Kbyte) and send (32 Kbyte) buffer capacity.  If for any
 * performance reason, more or less buffer capactiy is needed, this routine should be called prior to
 * any calls to SystemInit() or SystemCycle().  For example, VxWorks builds which do not have a large
 * amount of network resources might want to set these values smaller than the defaults.  Video system
 * clients expecting extreme amounts of incoming data might want to set these values to a much larger
 * amount.
 *
 * This call allows the setting of the server side receive buffer space as well as the client side
 * send buffer space. This is potentially applicable in rare cases (middle layer servers) where large
 * data payloads are sent as input from a client to a server.
 *
 * As in the case of _SystemAssignBufferSpace(), the UDP specific send buffer size is kept to a reasonable
 * value, no larger than twice the registered packet MTU.  UDP datagram transmission is not inhibited by
 * any client-server flow control which might cause the send buffer space to stack up.
 *
 * \param clnRcvBufferSpace is the amount of receive buffer space (in bytes) to assign to the configurable TINE
 * client sockets. A value of '0' will assign the default receive buffer capacity.
 * \param srvRcvBufferSpace is the amount of receive buffer space (in bytes) to assign to the configurable TINE
 * server sockets. A value of '0' will assign the same value as clnRcvBufferSpace (and damping this value by a
 * factor of 2 should it exceed the default receive buffer capacity).
 * \param clnSndBufferSpace is the amount of send buffer space (in bytes) to assign to the configurable TINE
 * client sockets. A value of '0' will assign the same value as srvSndBufferSpace (and damping this value by a
 * factor of 2 should it exceed the default receive buffer capacity).
 * \param srvSndBufferSpace is the amount of send buffer space (in bytes) to assign to the configurable
 * TINE server TCP or STREAM sockets.
 * The UDP socket buffer space will be at least as large as twice the assigned UDP packet MTU.
 * A value of 0' will assign the default send buffer capacity.
 *
 * \return 0 on success or a TINE error code
 *
 * \see SetPacketMTU(), _SystemAssignBufferSpace()
 */
TINE_EXPORT int _SystemAssignBufferSpaceEx(UINT32 clnRcvBufferSpace,UINT32 srvRcvBufferSpace,UINT32 clnSndBufferSpace,UINT32 srvSndBufferSpace);
#define SystemAssignBufferSpaceEx(a,b,c,d) _SystemAssignBufferSpaceEx(a,b,c,d)

/**
 * \brief Assigns explicit udp send buffer capacity
 *
 * The standard TINE client and server UDP sockets are initiated with a
 * reasonable default amount of send (32 Kbyte) buffer capacity.  If the UDP packet MTU is
 * set to a value greater than twice the existing UDP buffer capacity then it will be automatically
 * augmented to twice the MTU size. In the unlikely event that is becomes necessary to
 * set the UDP send buffer capacity explicity, it can be accomplished with this call.  As UDP
 * datagrams are not subject to client-server flow control they are generally transmitted as fast as
 * possible.  A datagram send operation is unlikely to fail under most circumstances,
 * unless a client or server has an unusually high number of simultaneous outstanding send
 * operations.
 *
 * \param sndBufferSpace is the amount of receive buffer space (in bytes) to assign to UDP sockets.
 *
 * \return 0 on success or a TINE error code
 *
 * \see SetPacketMTU(), _SystemAssignBufferSpace()
 */
TINE_EXPORT int _SystemAssignUdpSendBufferSpace(UINT32 sndBufferSpace);

/* SetUseLoopback() */
/**
 * \fn TINE_EXPORT void SetUseLoopback(int value)
 * \brief Sets the 'useLoopBack' flag to the value given (boolean).
 *
 * If the useLoopBack flag is 'TRUE' then client-side calls to a server on the same host as
 * the client will use the network socket layer as the communication avenue between client and
 * server, as opposed to shared memory (windows) or pipes (unix).
 *
 * \param value should be non-zero (TRUE) to turn this flag on or FALSE to turn it off.  FALSE is
 * the default.
 *
 * \see GetUseLoopback().
 */
TINE_EXPORT void SetUseLoopback(int value);
/* GetUseLoopback() */
/**
 * \fn TINE_EXPORT int GetUseLoopback(void)
 * \brief Returns the current value of the 'useLoopBack' flag.
 *
 * If the useLoopBack flag is 'TRUE' then client-side calls to a server on the same host as
 * the client will use the network socket layer as the communication avenue between client and
 * server, as opposed to shared memory (windows) or pipes (unix).
 *
 * \return the current setting of the useLoopBack flag.
 *
 * \see SetUseLoopback().
 */
TINE_EXPORT int GetUseLoopback(void);

/* SetBurstLimit() */
/**
 * \fn TINE_EXPORT int SetBurstLimit(int npackets)
 * \brief Sets the burst limit in number of packets which are allowed to be set consecutively
 *
 * For UDP transport (the default) TINE does not make use of flow control in any systematic way,
 * other than adjusting this parameter setting at the server side.  This parameter works together
 * with the cycle delay and determines how many UDP packets (datagrams) are allowed to be sent
 * consecutively within the transport loop before a brake is applies (cycle delay).  If all clients
 * and servers are on the same speed ethernet, this value can be adjusted to a much larger value.
 * However, a fast server sending to a slow client (a remote client via a DSL link for instance) might
 * send more datagrams in a burst than can be received.  The default value should compensate for this
 * effect.
 *
 * \param npackets is the number of packet allowed to be set consecutively for a given call
 *  (default: 1000).
 *
 * \return The current setting for the burst limit.
 *
 * \see SetCycleDelay(), GetBurstLimit().
 */
TINE_EXPORT int SetBurstLimit(int npackets);

/* GetBurstLimit() */
/**
 * \fn TINE_EXPORT int GetBurstLimit(void)
 * \brief Gets the burst limit in number of packets which are allowed to be sent consecutively
 *
 * For UDP transport (the default) TINE does not make use of flow control in any systematic way,
 * other than adjusting this parameter setting at the server side.  This parameter works together
 * with the cycle delay and determines how many UDP packets (datagrams) are allowed to be sent
 * consecutively within the transport loop before a brake is applies (cycle delay).  If all clients
 * and servers are on the same speed ethernet, this value can be adjusted to a much larger value.
 * However, a fast server sending to a slow client (a remote client via a DSL link for instance) might
 * send more datagrams in a burst than can be received.  The default value should compensate for this
 * effect.
 *
 * \return The current setting for the burst limit.
 *
 * \see SetBurstLimit(), SetCycleDelay().
 */
TINE_EXPORT int GetBurstLimit(void);

/* SetCycleDelay() */
/**
 * \fn TINE_EXPORT int SetCycleDelay(int msecs)
 * \brief Sets the delay time in milliseconds following a 'burst' of ethernet packets (as defined by
 * SetBurstLimit().
 *
 * For UDP transport (the default) TINE does not make use of flow control in any systematic way,
 * other than adjusting this parameter setting at the server side.  This parameter works together
 * with the burst limit and determines how long to delay further datagram transmission when the burst
 * limit has been reached. If all clients
 * and servers are on the same speed ethernet, this value can be adjusted to 0.
 * However, a fast server sending to a slow client (a remote client via a DSL link for instance) might
 * send more datagrams in a burst than can be received.  The default value should compensate for this
 * effect.
 *
 * \param msec is the time in milliseconds to delay before sending further ethernet packets.
 *  (default: 20 msec).
 *
 * \return The current setting for the cycle delay.
 *
 * \see SetBurstLimit().
 */
TINE_EXPORT int SetCycleDelay(int msecs);

/* SetCycleMicroDelay() */
/**
 * \fn TINE_EXPORT int SetCycleMicroDelay(int usecs)
 * \brief Sets the delay time in microseconds following a 'burst' of ethernet packets (as defined by
 * SetBurstLimit().
 *
 * For UDP transport (the default) TINE does not make use of flow control in any systematic way,
 * other than adjusting this parameter setting at the server side.  This parameter works together
 * with the burst limit and determines how long to delay further datagram transmission when the burst
 * limit has been reached. If all clients
 * and servers are on the same speed ethernet, this value can be left at its default value 0.
 * However, when fine adjustments in flow control are necessary, a delay in the micro second region
 * following each 'burst' of packets can be applied, where the delay in microseconds is given by
 * this setting.
 *
 * \param usec is the time in microseconds to delay before sending further ethernet packets.
 *  (default: 0 usec => use cycle delay in milliseconds).
 *
 * \return The current setting for the cycle micro delay.
 *
 * \see SetCycleDelay(), SetBurstLimit().
 */
TINE_EXPORT int SetCycleMicroDelay(int usecs);

/* GetCycleDelay() */
/**
 * \fn TINE_EXPORT int GetCycleDelay(void)
 * \brief Gets the delay time in milliseconds following a 'burst' of ethernet packets (as defined by
 * SetBurstLimit().
 *
 * For UDP transport (the default) TINE does not make use of flow control in any systematic way,
 * other than adjusting this parameter setting at the server side.  This parameter works together
 * with the burst limit and determines how long to delay further datagram transmission when the burst
 * limit has been reached. If all clients
 * and servers are on the same speed ethernet, this value can be adjusted to 0.
 * However, a fast server sending to a slow client (a remote client via a DSL link for instance) might
 * send more datagrams in a burst than can be received.  The default value should compensate for this
 * effect.
 *
 * \return The current setting for the cycle delay.
 *
 * \see SetBurstLimit(), SetCycleDelay().
 */
TINE_EXPORT int GetCycleDelay(void);

/* GetCycleMicroDelay() */
/**
 * \fn TINE_EXPORT int GetCycleMicroDelay(void)
 * \brief Gets the delay time in microseconds following a 'burst' of ethernet packets (as defined by
 * SetBurstLimit().
 *
 * For UDP transport (the default) TINE does not make use of flow control in any systematic way,
 * other than adjusting this parameter setting at the server side.  This parameter works together
 * with the burst limit and determines how long to delay further datagram transmission when the burst
 * limit has been reached. If all clients
 * and servers are on the same speed ethernet, this value can left at its default value 0.
 * However, when fine adjustments in flow control are necessary, a delay in the micro second region
 * following each 'burst' of packets can be applied, where the delay in microseconds is given by
 * this setting.
 *
 * \return The current setting for the cycle micro delay.
 *
 * \see SetBurstLimit(), SetCycleDelay().
 */
TINE_EXPORT int GetCycleMicroDelay(void);

/* SetForceSharedMemory() */
/**
 * \fn TINE_EXPORT void SetForceSharedMemory(int value)
 * \brief Disables or enables the 'useLoopBack' flag according to the value given (boolean).
 *
 * If the useLoopBack flag is 'FALSE' (default) then client-side calls to a server on the same host as
 * the client will attempt to use shared memory (windows) or pipes (unix) as a more expedient
 * communications avenue between client and server.  If the server however appears not to support
 * such communications (either a legacy server or a java server, for example), then the loopback address
 * is used as a fallback.  If the server is known a priori to support such alternate communication
 * channels, then this 'fallback' strategy can be turned off by making use of this call.
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.  FALSE is
 * the default.
 *
 * \see GetForceSharedMemory().
 */
TINE_EXPORT void SetForceSharedMemory(int value);
/* GetForceSharedMemory() */
/**
 * \fn TINE_EXPORT int GetForceSharedMemory(void)
 * \brief returns the current setting of this feature.
 *
 * If the useLoopBack flag is 'FALSE' (default) then client-side calls to a server on the same host as
 * the client will attempt to use shared memory (windows) or pipes (unix) as a more expedient
 * communications avenue between client and server.  If the server however appears not to support
 * such communications (either a legacy server or a java server, for example), then the loopback address
 * is used as a fallback.  If the server is known a priori to support such alternate communication
 * channels, then this 'fallback' strategy can be turned off by making use of SetForceSharedMemory().
 * The current setting can be seen by using this call.
 *
 * \return the current setting of the forcedSharedMemory feature.
 *
 * \see SetForceSharedMemory().
 */
TINE_EXPORT int GetForceSharedMemory(void);

/**
 * \fn TINE_EXPORT void SetTcpConnectionTableSize(int value)
 * \brief sets the tcp connection table size.
 *
 * Transport communication goes via UDP by default.  However by specifying
 * either CM_CONNECT or CM_STREAM a caller can request connected stream
 * transport via TCP.  In this case a slot is reserved in a tcp connection
 * managment table for the resulting communications socket.  This is a
 * table which is NOT allowed to grow in size but is restricted to the
 * size seen at initialization (default = 32 entries).  If it is known
 * a priori that TCP communication will be used in a big way, then this
 * table size should be adjusted prior to system initialization
 * (i.e. a call to SystemInit()).
 *
 * \param value is the desired tcp connection table size (minimum 8)
 *
 * \see GetTcpConnectionTableSize().
 */
TINE_EXPORT void SetTcpConnectionTableSize(int value);

/**
 * \fn TINE_EXPORT int GetTcpConnectionTableSize(void)
 * \brief returns the current setting of the tcp connection table size.
 *
 * Transport communication goes via UDP by default.  However by specifying
 * either CM_CONNECT or CM_STREAM a caller can request connected stream
 * transport via TCP.  In this case a slot is reserved in a tcp connection
 * managment table for the resulting communications socket.  This is a
 * table which is NOT allowed to grow in size but is restricted to the
 * size seen at initialization (default = 32 entries).
 *
 * \return the current setting of the tcp connection table size.
 *
 * \see SetTcpConnectionTableSize().
 */
TINE_EXPORT int GetTcpConnectionTableSize(void);

/**
 * \fn TINE_EXPORT void SetIdleConnectionDuration(int allowedIdleSeconds)
 * \brief sets the time-to-live for an unused TCP connection to a server
 *
 * A client might choose to make use of TCP transport for a short period
 * of time (e.g. a transation) and then no longer require TCP communications
 * to the target in question.  In order to conserve resources in the long run
 * the client side will close such unused connections after the
 * allowed idle time has expired.
 *
 * Use this routine to set this value (default = 300 seconds)
 *
 * \param allowedIdleSeconds is the allowed TCP connection idle time in
 * seconds.
 */
TINE_EXPORT void SetIdleConnectionDuration(int allowedIdleSeconds);

/**
 * \fn TINE_EXPORT void SetAllowRemoteDebugging(int value)
 * \brief enables/disables remote debugging
 *
 * The 'attachfec' tool is usually used on the same host as
 * the running server, in which case the debugging output is
 * forwarding via an efficient local named pipe.
 * In some cases, it might be desired to make use of the 'attachfec' on a remote
 * host, in which case debugging output is transfered via a stream
 * connection (less efficient).  This feature can be turned OFF by
 * making use of this API call and passing a value of FALSE prior
 * to a call to SystemInit().
 *
 * \param value is the desired setting (default = TRUE).
 *
 * \see GetAllowRemoteDebugging().
 */
TINE_EXPORT void SetAllowRemoteDebugging(int value);

/**
 * \fn TINE_EXPORT int GetAllowRemoteDebugging(void)
 * \brief returns the current setting of this value
 *
 * The 'attachfec' tool is usually used on the same host as
 * the running server, in which case the debugging output is
 * forwarding via an efficient local named pipe.
 * In some cases, it might be desired to make use of the 'attachfec' on a remote
 * host, in which case debugging output is transfered via a stream
 * connection (less efficient).  Whether this feature is available is
 * determined by this setting.
 *
 * \return the current setting (default = TRUE).
 *
 * \see SetAllowRemoteDebugging().
 */
TINE_EXPORT int GetAllowRemoteDebugging(void);

TINE_EXPORT void SetSendGlobalsAsBroadcastOnly(int value);

TINE_EXPORT void SetTcpBurstLimit(int value);
TINE_EXPORT int GetTcpBurstLimit(void);
extern int streamQueueOverwriteEager;
TINE_EXPORT void SetStreamQueueOverwriteEager(int value);
TINE_EXPORT int GetStreamQueueOverwriteEager(void);

#if !defined(VXWORKS) && !defined(VMS) && !defined(SMALL_TINE_LIB)
void initAddrPortsFromCsv(void);
void initAddrPortsFromEnv(void);
extern int gInitAddrPortsDone;
#else
# ifndef DEFAULT_TRANSPORT_THREAD
#   define DEFAULT_TRANSPORT_THREAD FALSE
# endif
#endif
void initSockBufferSizeFromEnv(void);
void initIpFromEnv(void);

extern int gListenForGlobals;
extern int gWaitingOnSelect;
#if defined(VXWORKS) && (_WRS_VXWORKS_MAJOR < 6)
struct hostent *gethostbyname(const char *hstname);
struct hostent *gethostbyaddr(const char *hstaddr, int hstlen, int hsttype);
#endif
#ifdef WIN32
int GetErrno(void);
int WinsockLoaded(void);
# ifndef EISCONN
#   define EISCONN WSAEISCONN
# endif
# ifndef EWOULDBLOCK
#   define EWOULDBLOCK WSAEWOULDBLOCK
# endif
# ifndef ENOBUFS
#   define ENOBUFS WSAENOBUFS
# endif
#endif /* WIN32 defined */
#ifdef UNIX
typedef struct
{
  int sck;
  int fecIdx;
} ClnPipeSck;
# define PIPE_TBL_CAPACITY 100
extern ClnPipeSck gClnPipeSckTbl[PIPE_TBL_CAPACITY];
extern int nClnPipeSck;
int addClnPipeSck(int sck,int idx);
int rmvClnPipeSck(int sck);
typedef struct
{
  SOCKET sck;
  int type;
  char name[16];
  SCKADR ipAddr;
} IPCSCKT;
extern char gIpcTinePipe[64];
extern int nIPCsck;
extern IPCSCKT *ipcSckTbl;
extern int ipcConTblSize;
int isUnixSocket(int sck);
void ClosePipeSocket(int sck);
# define writeto(fd,bf,sz,fg,ad,ln) (\
  !isUnixSocket(fd) ? sendto(fd,bf,sz,fg,ad,ln) : write(fd,bf,sz)\
)
# define readfrom(fd,bf,sz,fg,ad,ln) (\
  !isUnixSocket(fd) ? recvfrom(fd,bf,sz,fg,ad,(_socklen_t *)ln) : read(fd,bf,sz)\
)
# define getpeeraddr(fd,ad,ln) getpeername(fd,ad,(_socklen_t *)ln)
#elif defined(VXWORKS)
# define writeto(fd,bf,sz,fg,ad,ln) (IsConnProt(inetProtocol) ? send(fd,bf,sz,fg) : sendto(fd,bf,sz,fg,ad,ln))
# define readfrom(fd,bf,sz,fg,ad,ln) recvfrom(fd,bf,sz,fg,ad,ln)
# define getpeeraddr(fd,ad,ln) getpeername(fd,ad,(int *)ln)
#else /* !UNIX and !VXWORKS */
# define writeto(fd,bf,sz,fg,ad,ln) sendto(fd,bf,sz,fg,ad,ln)
# define readfrom(fd,bf,sz,fg,ad,ln) recvfrom(fd,bf,sz,fg,ad,ln)
# define getpeeraddr(fd,ad,ln) getpeername(fd,ad,(int *)ln)
#endif /* UNIX defined or VXWORKS defined */

#if !defined(MSG_NOSIGNAL)
# define MSG_NOSIGNAL 0
#endif

int processTCPDataStream(TCPBCKT **pbckt,int nread);
void recvIncomingIPDataErr(SOCKET sck,int inetprot,int nread);
int processTCPDataPacket(SOCKET sck,BYTE *bf,int nread,SCKADR sckadr,int inetprot);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* SRVIP_H_ */
